/***************************************************************************
*   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
*   rvinyard@cs.nmsu.edu                                                  *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU Lesser General Public License as        *
*   published by the Free Software Foundation version 2.1.                *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU Lesser General Public      *
*   License along with this library; if not, write to the                 *
*   Free Software Foundation, Inc.,                                       *
*   51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA              *
***************************************************************************/
#include "fieldbuffer.h"

#include "fieldvector.h"
#include "recordbuffer.h"
#include "utility.h"

#include <algorithm>

namespace bit
  {

  FieldBuffer::FieldBuffer(RecordBuffer& b, FieldBase::pointer f): m_buffer(b), m_field(f)
  { }

  FieldBuffer::~FieldBuffer()
  { }

  const std::string & FieldBuffer::xml( )
  {
    std::ostringstream sout;
    Data d;

    sout << "<data name=\"" << m_field->name();

    if ( m_field->container_type() == FIELD_ELEMENT )
      sout << "[" << bit_dynamic_pointer_cast<FieldVector::Element>(m_field)->index() << "]";

    sout << "\">";

    switch ( m_field->type().type() ) {
      default:
        d = this->data();
        sout << "0x" << d.hex_string();
    }

    sout << "</data>";

    m_xml = sout.str();
    return m_xml;
  }

  FieldBase::pointer FieldBuffer::field() {
    return m_field;
  }

  RecordBuffer& FieldBuffer::buffer() {
    return m_buffer;
  }

  FieldBuffer FieldBuffer::field(size_t index) throw (exception::invalid_index)
  {
    if (m_field)
      return FieldBuffer(m_buffer, m_field->field(index) );
    else
      throw exception::invalid_index();
  }

  FieldBuffer FieldBuffer::field(std::string index) throw (exception::invalid_index)
  {
    if (m_field)
      return FieldBuffer(m_buffer, m_field->field(index) );
    else
      throw exception::invalid_index();
  }

  FieldBuffer FieldBuffer::operator[](size_t index) throw (exception::invalid_index)
  {
    return this->field(index);
  }

  FieldBuffer FieldBuffer::operator[](std::string index) throw (exception::invalid_index)
  {
    return this->field(index);
  }


  Data FieldBuffer::data( )
  {
    Data d = Data( m_field->length(OCTETS) );
    this->unpack( d.data(), m_field->length(OCTETS) );
    return d;
  }

  bool FieldBuffer::unpack( void * mem, size_t mem_octets )
  {
    return m_buffer.unpack(mem, mem_octets, END, m_field->start(BITS), m_field->length(BITS));
  }

  bool FieldBuffer::pack( const void * mem, size_t mem_octets )
  {
    bool b = m_buffer.pack(mem, mem_octets, END, m_field->start(BITS), m_field->length(BITS), m_field->length(BITS));
    if (b)
      //     m_buffer.on_field_changed(m_field->full_name(m_field->depth()-1));
      m_buffer.on_field_changed( m_field );
    return b;
  }

  bool FieldBuffer::pack( const void * mem, size_t mem_octets, size_t n )
  {
    bool b = m_buffer.pack(mem, mem_octets, END, m_field->start(BITS), m_field->length(BITS), n);
    if (b)
      //     m_buffer.on_field_changed(m_field->full_name(m_field->depth()-1));
      m_buffer.on_field_changed( m_field );
    return b;
  }

  bool FieldBuffer::unpack_uint64( uint64_t& ui64 )
  {
    ui64=0x00;

    // ensure that length is no more than 64
    size_t length = std::min( (size_t)64, m_field->length(BITS));

//     std::cout << "Unpacking " << m_field->name() << " with byte order " << m_field->type().byte_order() << std::endl;

    if ( m_field->type().byte_order() == BYTEORDER_LITTLE_ENDIAN ) {
      if ( ! m_buffer.unpack(&ui64, sizeof(uint64_t), START, m_field->start(BITS), length ) )
        return false;
      // need to expand upper octet
      if ( length%8 != 0) {
        uint8_t* upper_octet = (uint8_t*)(&ui64) + length/8;
        *upper_octet >>= 8 - length%8;
      }
      ui64 = le_to_host(ui64);
    }
    else {
      if ( ! m_buffer.unpack(&ui64, sizeof(uint64_t), END, m_field->start(BITS), length ) )
        return false;
      if ( m_field->type().byte_order() == BYTEORDER_NETWORK )
        ui64 = net_to_host(ui64);
      else if ( m_field->type().byte_order() == BYTEORDER_BIG_ENDIAN )
        ui64 = be_to_host(ui64);
    }
    return true;
  }

  bool FieldBuffer::pack_uint64(uint64_t ui64){
    bool b;

    // ensure that length is no more than 64
    size_t length = std::min( (size_t)64, m_field->length(BITS));

    if ( m_field->type().byte_order() == BYTEORDER_LITTLE_ENDIAN ) {
      ui64 = host_to_le(ui64);
      // need to compress upper octet
      if ( length%8 != 0) {
        uint8_t* upper_octet = (uint8_t*)(&ui64) + length/8;
        *upper_octet <<= 8 - length%8;
      }
      b = m_buffer.pack(&ui64, sizeof(uint64_t), START, m_field->start(BITS), length, length);
    }
    else {
      if ( m_field->type().byte_order() == BYTEORDER_NETWORK )
        ui64 = host_to_net(ui64);
      else if ( m_field->type().byte_order() == BYTEORDER_BIG_ENDIAN )
        ui64 = host_to_be(ui64);
      b = m_buffer.pack(&ui64, sizeof(uint64_t), END, m_field->start(BITS), length, length);
    }

    if (b)
      //     m_buffer.on_field_changed(m_field->full_name(m_field->depth()-1));
      m_buffer.on_field_changed( m_field );
    return b;
  }

  bool FieldBuffer::unpack_float(float& f) {
    bool b = unpack(&f, sizeof(float));
    switch(m_field->type().byte_order()) {
      case BYTEORDER_HOST:
        break;
      case BYTEORDER_NETWORK:
        f = net_to_host(f);
        break;
      case BYTEORDER_BIG_ENDIAN:
        f = be_to_host(f);
        break;
      case BYTEORDER_LITTLE_ENDIAN:
        f = le_to_host(f);
        break;
    }
    return b;
  }

  bool FieldBuffer::unpack_double(double& d) {
    bool b = unpack(&d, sizeof(double));
    switch(m_field->type().byte_order()) {
      case BYTEORDER_HOST:
        break;
      case BYTEORDER_NETWORK:
        d = net_to_host(d);
        break;
      case BYTEORDER_BIG_ENDIAN:
        d = be_to_host(d);
        break;
      case BYTEORDER_LITTLE_ENDIAN:
        d = le_to_host(d);
        break;
    }
    return b;
  }

  bool FieldBuffer::pack_double(double d) {
    // TODO put in proper throw
    if ( m_field->length(BITS) != 64) throw;
    switch(m_field->type().byte_order()) {
      case BYTEORDER_HOST:
        break;
      case BYTEORDER_NETWORK:
        d = host_to_net(d);
        break;
      case BYTEORDER_BIG_ENDIAN:
        d = host_to_be(d);
        break;
      case BYTEORDER_LITTLE_ENDIAN:
        d = host_to_le(d);
        break;
    }
    return pack(&d, sizeof(double));
  }

  bool FieldBuffer::pack_float(float f) {
    // TODO put in proper throw
    if (m_field->length(BITS) != 32) throw;
    switch(m_field->type().byte_order()) {
      case BYTEORDER_HOST:
        break;
      case BYTEORDER_NETWORK:
        f = host_to_net(f);
        break;
      case BYTEORDER_BIG_ENDIAN:
        f = host_to_be(f);
        break;
      case BYTEORDER_LITTLE_ENDIAN:
        f = host_to_le(f);
        break;
    }
    return pack(&f, sizeof(float));
  }

}
