/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation version 2.1.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the                 *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA              *
 ***************************************************************************/
#include "fieldvector.h"

#include <iostream>

#include <string>
#include <sstream>
#include <ctype.h>

#define WHOLE_UNITS(__offset, __units) \
  ( ((__offset) > 0.0) ? ( (__offset)/__units+(((__offset)%__units)?1:0) ) : ( (__offset)/__units+(((__offset)%__units)?(-1):0)) )

namespace bit
  {

  FieldVector::Element::Element(size_t index): m_index(index)
  {}

  FieldVector::Element::pointer FieldVector::Element::create(size_t index)
  {
    return pointer( new Element(index) );
  }

  const std::string & FieldVector::xml( )
  {
    std::ostringstream sout;
    sout << "<field name=\"" << m_name
        << "\" start=\"" << m_offset
        << "\" start_units=\"" << m_offset_units
        << "\" length=\"" << m_length
        << "\" length_units=\"" << m_length_units
        << "\" vector_size=\"" << m_size << "\"";
    if ( m_type.type() == TYPE_NONE )
      sout << "/>";
    else
      sout << ">" << m_type.xml() << "</field>";
    m_xml = sout.str();
    return m_xml;
  }

  size_t FieldVector::Element::length() const
    {
      return dynamic_cast<FieldVector*>(m_parent)->element_length();
    }

  size_t FieldVector::Element::length(size_t units) const
    {
      return FieldBase::length(units);
    }

  size_t FieldVector::Element::length_units() const
    {
      return m_parent->length_units();
    }

  int FieldVector::Element::offset() const
    {
      return m_parent->offset() + dynamic_cast<FieldVector*>(m_parent)->element_length() * m_index;
    }

  int FieldVector::Element::offset(size_t units) const
    {
      return FieldBase::offset(units);
    }

  size_t FieldVector::Element::offset_units() const
    {
      return m_parent->length_units();
    }

  std::string FieldVector::Element::name() const
    {
      std::ostringstream sout;
      sout << m_parent->name() << "[" << m_index << "]";
      return sout.str();
    }

  std::string FieldVector::Element::name(int depth) const
    {
      return this->FieldBase::name(depth);
    }

  std::string FieldVector::Element::description() const
    {
      return m_parent->description();
    }

  size_t FieldVector::Element::index() const
    {
      return m_index;
    }

    const std::string & FieldVector::Element::xml( )
    {
      std::ostringstream sout;
      sout << "<field name=\"" << this->name()
          << "\" start=\"" << this->offset()
          << "\" start_units=\"" << this->offset_units()
          << "\" length=\"" << this->length()
          << "\" length_units=\"" << this->length_units() << "\"";
      if ( m_type.type() == TYPE_NONE )
        sout << "/>";
      else
        sout << ">" << m_type.xml() << "</field>";
      m_xml = sout.str();
      return m_xml;
    }

  FieldBase::pointer FieldVector::Element::clone()
  {
    Element::pointer element = Element::create(m_index);
    return element;
  }


  FieldVector::FieldVector( std::string name,
                            size_t size,
                            size_t element_length,
                            size_t element_length_units,
                            int offset,
                            size_t offset_units,
                            std::string description
                          ):
      Field(name, element_length, element_length_units, offset, offset_units, description),
      m_size(0)
  {
    set_size(size);
  }

  FieldVector::pointer FieldVector::create( std::string name,
      size_t size,
      size_t element_length,
      size_t element_length_units,
      int offset,
      size_t offset_units,
      std::string description
                                          )
  {
    return FieldVector::pointer ( new FieldVector(name, size, element_length, element_length_units, offset, offset_units, description) );
  }

  FieldVector::~FieldVector()
  { }

  size_t FieldVector::length() const
    {
      return m_length * m_size;
    }

  size_t FieldVector::length(size_t units) const
    {
      return FieldBase::length(units);
    }

  size_t FieldVector::element_length() const
    {
      return m_length;
    }

  size_t FieldVector::size() const
    {
      return m_size;
    }

  void FieldVector::set_size(size_t size)
  {
    Element::pointer element;

    if (size > m_size)
      for (size_t i=m_size; i < size; i++)
        {
          element = Element::create(i);
          element->m_parent = this;
          m_elements.push_back( element );
        }
    else if (size < m_size)
      m_elements.erase(m_elements.begin()+size, m_elements.end());
    m_size = size;
  }

  FieldBase::iterator FieldVector::begin()
  {
    if (m_size > 0)
      return FieldBase::iterator(this, m_elements[0] );
    else
      return this->end();
  }

  FieldBase::iterator FieldVector::end()
  {
    return FieldBase::iterator(this, FieldVector::Element::pointer() );
  }

  FieldBase::pointer FieldVector::field(size_t i)
  {
    if (i < m_size)
      return m_elements[i];
    else
      return FieldBase::pointer();
  }

  FieldBase::pointer FieldVector::previous_field(FieldBase::pointer current_field) throw (exception::invalid_container_op)
  {
    FieldVector::Element::pointer element;

    if (!current_field)
      return FieldVector::Element::pointer();

    try
      {
        element = bit_dynamic_pointer_cast<FieldVector::Element>(current_field);
      }
    catch ( std::bad_cast )
      {
        throw exception::invalid_container_op();
      }

    if (element->index() > 0)
      return m_elements[ element->index()-1 ];
    else
      return FieldVector::Element::pointer();
  }

  FieldBase::pointer FieldVector::next_field(FieldBase::pointer current_field) throw (exception::invalid_container_op)
  {
    FieldVector::Element::pointer element;

    if (!current_field)
      return FieldVector::Element::pointer();

    try
      {
        element = bit_dynamic_pointer_cast<FieldVector::Element>(current_field);
      }
    catch ( std::bad_cast )
      {
        throw exception::invalid_container_op();
      }

    if (element->index()+1 < m_size)
      return m_elements[ element->index()+1 ];
    else
      return FieldVector::Element::pointer();
  }

  FieldBase::pointer FieldVector::clone()
  {
    FieldVector::pointer field = FieldVector::create(m_name, m_size, m_length, m_length_units, m_offset, m_offset_units, m_description);
    return field;
  }


}
