/***************************************************************************
*   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
*   rvinyard@cs.nmsu.edu                                                  *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU Lesser General Public License as        *
*   published by the Free Software Foundation version 2.1.                *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU Lesser General Public      *
*   License along with this library; if not, write to the                 *
*   Free Software Foundation, Inc.,                                       *
*   51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA              *
***************************************************************************/
#include "recordvector.h"

#include <iostream>

#define WHOLE_UNITS(__offset, __units) \
  ( ((__offset) > 0.0) ? ( (__offset)/__units+(((__offset)%__units)?1:0) ) : ( (__offset)/__units+(((__offset)%__units)?(-1):0)) )

namespace bit
  {

  RecordVector::Element::Element( size_t index, RecordStorage& reference_layout ):
      m_index(index),
      m_fields(*this)
  {
    m_fields = reference_layout;
  }

  RecordVector::Element::pointer RecordVector::Element::create( size_t index, RecordStorage& reference_layout )
  {
    return Element::pointer( new Element(index, reference_layout) );
  }

  RecordVector::Element::~Element()
  { }

  size_t RecordVector::Element::length() const
    {
      size_t length = 0;
      RecordStorage::Fields::const_iterator i;

      for (i = m_fields.field_set.begin(); i != m_fields.field_set.end(); i++)
      {
      // get the length of each child as its internal offset + its length in the length_units of this record
        size_t bits = (*i)->offset(BITS) + (*i)->length(BITS);
        length = std::max(length, bits);
      }

      return WHOLE_UNITS(length, this->length_units() );
    }

  size_t RecordVector::Element::length(size_t units) const
    {
      return this->FieldBase::length(units);
    }

  size_t RecordVector::Element::length_units() const
    {
      return m_parent->length_units();
    }

  int RecordVector::Element::offset() const
    {
      return (this->length() * m_index);
    }

  int RecordVector::Element::offset(size_t units) const
    {
      return this->FieldBase::offset( units );
    }

  size_t RecordVector::Element::offset_units() const
    {
      return m_parent->offset_units();
    }

  std::string RecordVector::Element::name() const
    {
      std::ostringstream sout;
      sout << m_parent->name() << "[" << m_index << "]";
      return sout.str();
    }

  std::string RecordVector::Element::name(int depth) const
    {
      return FieldBase::name(depth);
    }

  std::string RecordVector::Element::description() const
    {
      return m_parent->description();
    }

    const std::string & RecordVector::xml( )
    {
      std::ostringstream sout;
      sout << "<record name=\"" << m_name
          << "\" length_units=\"" << m_length_units
          << "\" start=\"" << m_offset
          << "\" start_units=\"" << m_offset_units
          << "\" vector_size=\"" << m_size << "\">\n";

      if ( m_type.type() != TYPE_NONE )
        sout << m_type.xml() << "\n";

      iterator i;
      for ( i = m_record_layout.begin(); i != m_record_layout.end(); i++ )
        sout << i->xml() << "\n";

      sout << "</record>\n";
      m_xml = sout.str();
      return m_xml;
    }

    FieldBase::iterator RecordVector::Element::begin()
  {
    return m_fields.begin();
  }

  FieldBase::iterator RecordVector::Element::end()
  {
    return m_fields.end();
  }

  FieldBase::pointer RecordVector::Element::field(std::string s) {
    return m_fields.field(s);
  }

  FieldBase::pointer RecordVector::Element::field(size_t i) {
    return m_fields.field(i);
  }

  size_t RecordVector::Element::fields()
  {
    return m_fields.fields();
  }

  size_t RecordVector::Element::index()
  {
    return m_index;
  }

  FieldBase::pointer RecordVector::Element::clone() {
    return Element::create(m_index, m_fields);
  }

  FieldBase::pointer RecordVector::Element::previous_field(FieldBase::pointer current_field) throw (exception::invalid_container_op) {
    return m_fields.previous_field(current_field);
  }

  FieldBase::pointer RecordVector::Element::next_field(FieldBase::pointer current_field) throw (exception::invalid_container_op) {
    return m_fields.next_field(current_field);
  }




  RecordVector::RecordVector( std::string name,
                              size_t size,
                              size_t length_units,
                              int offset,
                              size_t offset_units,
                              std::string descr
                            ):
      RecordBase(name, length_units, offset, offset_units, descr),
      m_record_layout(*this),
      m_size(0)
  {
    set_size(size);
  }

  RecordVector::RecordVector( const RecordStorage& layout,
                              std::string name,
                              size_t size,
                              size_t length_units,
                              int offset,
                              size_t offset_units,
                              std::string descr
                            ):
      RecordBase(name, length_units, offset, offset_units, descr),
      m_record_layout(*this),
      m_size(0)
  {
    m_record_layout = layout;
    set_size(size);
  }

  RecordVector::pointer RecordVector::create( std::string name,
                                              size_t size,
                                              size_t length_units,
                                              int offset,
                                              size_t offset_units,
                                              std::string descr
                                            )
  {
    return RecordVector::pointer( new RecordVector( name, size, length_units, offset, offset_units, descr ) );
  }

  RecordVector::pointer RecordVector::create( const RecordStorage& layout,
                                              std::string name,
                                              size_t size,
                                              size_t length_units,
                                              int offset,
                                              size_t offset_units,
                                              std::string descr
                                            )
  {
    return RecordVector::pointer( new RecordVector( layout, name, size, length_units, offset, offset_units, descr ) );
  }

  RecordVector::~RecordVector()
  { }

  size_t RecordVector::length( ) const
  {
    size_t length = 0;
    RecordStorage::Fields::const_iterator i;

    for (i = m_record_layout.field_set.begin(); i != m_record_layout.field_set.end(); i++)
    {
      // get the length of each child as its internal offset + its length in the length_units of this record
      size_t bits = (*i)->offset(BITS) + (*i)->length(BITS);
      length = std::max(length, bits);
    }

    length *= m_size;

    return WHOLE_UNITS(length, m_length_units);
  }

  size_t RecordVector::length( size_t units ) const
  {
    return this->FieldBase::length(units);
  }

  size_t RecordVector::size() const
    {
      return m_size;
    }

  void RecordVector::set_size(size_t size)
  {
    Element::pointer element;

    if ( size > m_elements.size() )
      {
        for (size_t i=m_elements.size(); i < size; i++) {
          element = Element::create(i, m_record_layout);
          element->m_parent = this;
          m_elements.push_back( element );
        }
      }
    else if ( size < m_elements.size() )
      {
        m_elements.erase(m_elements.begin()+size, m_elements.end());
      }

    m_size = size;
  }

  FieldBase::iterator RecordVector::begin()
  {
    if (m_elements.size() > 0)
      return FieldBase::iterator(this, m_elements[0] );
    else
      return this->end();
  }

  FieldBase::iterator RecordVector::end()
  {
    return FieldBase::iterator(this, RecordVector::Element::pointer() );
  }

  FieldBase::pointer RecordVector::field(size_t i)
  {
    if ( i < m_elements.size() )
      return m_elements[i];
    else
      return FieldBase::pointer();
  }

  void RecordVector::add_field(FieldBase::pointer child)
  {
    m_record_layout.add_field( child->clone() );
    for (size_t i = 0; i < m_size; i++)
      m_elements[i]->m_fields.add_field( child->clone() );
  }

  void RecordVector::remove_field(const std::string& name)
  {
    m_record_layout.remove_field( name );
    for (size_t i = 0; i < m_elements.size(); i++)
      m_elements[i]->m_fields.remove_field( name );
  }

  size_t RecordVector::fields()
  {
    return m_size;
  }

  FieldBase::pointer RecordVector::clone()
  {
    return RecordVector::create(m_record_layout, m_name, m_size, m_length_units, m_offset, m_offset_units, m_description);
  }

  FieldBase::pointer RecordVector::previous_child(FieldBase::pointer current_child) throw (exception::invalid_container_op)
  {
    RecordVector::Element::pointer element;

    if (!current_child)
      return RecordVector::Element::pointer();

    try
      {
        element = bit_dynamic_pointer_cast<RecordVector::Element>(current_child);
      }
    catch ( std::bad_cast )
      {
        throw exception::invalid_container_op();
      }

    if (element->index() > 0)
      return m_elements[ element->index()-1 ];
    else
      return RecordVector::Element::pointer();
  }

  FieldBase::pointer RecordVector::next_child(FieldBase::pointer current_child) throw (exception::invalid_container_op)
  {
    RecordVector::Element::pointer element;

    if (!current_child)
      return RecordVector::Element::pointer();

    try
      {
        element = bit_dynamic_pointer_cast<RecordVector::Element>(current_child);
      }
    catch ( std::bad_cast )
      {
        throw exception::invalid_container_op();
      }

    if ( element->index()+1 < m_elements.size() )
      return m_elements[ element->index()+1 ];
    else
      return RecordVector::Element::pointer();
  }


}
